@@ -23,29 +23,50 @@ class _Types:
     NONETYPE = type(None)
 
 
+# the input and output of TypeDAL.define
+T = typing.TypeVar("T", typing.Type["TypedTable"], typing.Type["Table"])
+
+
 class TypeDAL(pydal.DAL):
     """@DynamicAttrs"""
+
     dal: Table
 
     default_kwargs = {
-        # more risky to enable by default, so false:
-        'notnull': False,
-        # fields are required by default and can be nullable/optional with typing.Optional or Union with None (| None)
-        'required': True
+        # fields are 'required' (notnull) by default:
+        "notnull": True,
     }
 
-    def define(self, cls: typing.Type["TypedTable"]):
-        tablename = self._to_snake(cls.__name__)
-        table = self.define_table(
-            tablename,
-            *[self._to_field(fname, ftype)
+    def define(self, cls: T) -> T:
+        # when __future__.annotations is implemented, cls.__annotations__ will not work anymore as below.
+        # proper way to handle this would be (but gives error right now due to Table implementing magic methods):
+        # typing.get_type_hints(cls, globalns=None, localns=None)
 
-              for fname, ftype in cls.__annotations__.items()
-              ]
-        )
+        # dirty way (with evil eval):
+        # [eval(v) for k, v in cls.__annotations__.items()]
+        # this however also stops working when variables outside this scope or even references to other
+        # objects are used. So for now, this package will NOT work when from __future__ import annotations is used,
+        # and might break in the future, when this annotations behavior is enabled by default.
+        # todo: add to caveats
+
+        # non-annotated variables have to be passed to define_table as kwargs
+
+        tablename = self._to_snake(cls.__name__)
+        fields = [
+            self._to_field(fname, ftype) for fname, ftype in cls.__annotations__.items()
+        ]
+        other_kwargs = {
+            k: v
+            for k, v in cls.__dict__.items()
+            if k not in cls.__annotations__ and not k.startswith("_")
+        }
+
+        table = self.define_table(tablename, *fields, **other_kwargs)
 
         cls.__set_internals__(db=self, table=table)
 
+        # the ACTUAL output is not TypedTable but rather pydal.Table
+        # but telling the editor it is T helps with hinting.
         return table
 
     def __call__(self, *args, **kwargs) -> pydal.objects.Set:
@@ -65,7 +86,7 @@ class TypeDAL(pydal.DAL):
         return Field(name, type, **{**cls.default_kwargs, **kw})
 
     @classmethod
-    def _to_field(cls, fname, ftype, **kw):
+    def _to_field(cls, fname: str, ftype: type, **kw):
         fname = cls._to_snake(fname)
 
         if mapping := BASIC_MAPPINGS.get(ftype):
@@ -85,15 +106,16 @@ class TypeDAL(pydal.DAL):
             # list[...]
             _childtype = TypedFieldType._convert_generic_alias_list(ftype)
             return cls._build_field(fname, f"list:{_childtype}", **kw)
-        elif isinstance(ftype, typing._UnionGenericAlias) or isinstance(ftype, types.UnionType):
+        elif isinstance(ftype, typing._UnionGenericAlias) or isinstance(
+            ftype, types.UnionType
+        ):
             # typing.Optional[type] == type | None
             match ftype.__args__:
                 case (_child_type, _Types.NONETYPE):
                     # good union
 
-                    # if a field is optional, it can't be required or nutnull!
-                    kw['required'] = False
-                    kw['notnull'] = False
+                    # if a field is optional, it is nullable:
+                    kw["notnull"] = False
                     return cls._to_field(fname, _child_type, **kw)
                 case other:
                     raise NotImplementedError(f"Invalid type union '{other}'")
@@ -104,7 +126,9 @@ class TypeDAL(pydal.DAL):
     @staticmethod
     def _to_snake(camel: str) -> str:
         # https://stackoverflow.com/a/44969381
-        return ''.join(['_' + c.lower() if c.isupper() else c for c in camel]).lstrip('_')
+        return "".join(["_" + c.lower() if c.isupper() else c for c in camel]).lstrip(
+            "_"
+        )
 
 
 class TypedTableMeta(type):
@@ -129,7 +153,8 @@ class TypedTable(Table, metaclass=TypedTableMeta):
     @classmethod
     def __get_table_column__(cls, col):
         # db.table.col -> SomeTypedTable.col (via TypedTableMeta.__getattr__)
-        return cls.__table[col]
+        if cls.__table:
+            return cls.__table[col]
 
     def __new__(cls, *a, **kw):
         # when e.g. Table(id=0) is called without db.define,
@@ -159,8 +184,8 @@ class TypedFieldType(Field):
         return f"<{s} with options {self.kwargs}>"
 
     def __str__(self):
-        if 'type' in self.kwargs:
-            t = self.kwargs['type']
+        if "type" in self.kwargs:
+            t = self.kwargs["type"]
         else:
             t = self.type.__name__ if issubclass(type(self.type), type) else self.type
         return f"TypedField.{t}"
@@ -168,8 +193,8 @@ class TypedFieldType(Field):
     def _to_field(self, name: str, **extra_kwargs) -> Field:
         other_kwargs = self.kwargs.copy()
         other_kwargs.update(extra_kwargs)
-        if 'type' in other_kwargs:
-            _type = other_kwargs.pop('type')
+        if "type" in other_kwargs:
+            _type = other_kwargs.pop("type")
         else:
             _type = self._to_field_type(self.type)
 
@@ -207,10 +232,13 @@ class TypedFieldType(Field):
         return cls._to_field_type(childtype)
 
 
-T = typing.TypeVar('T')
+S = typing.TypeVar("S")
 
 
-def TypedField(_type: T, **kwargs) -> T:
-    # sneaky: het is een functie en geen class opdat er een return type is :)
-    # en de return type (T) is de input type in _type
-    return TypedFieldType(_type, **kwargs)
+class TypedRows(typing.Collection[S], Rows):
+    """
+    Can be used as the return type of a .select()
+    e.g.
+
+    people: TypedRows[Person] = db(Person).select()
+    """
